#ifndef MATRIX_H_
#define MATRIX_H_

#include <string>
#include <sstream>

#include "exception/GSException.h"

template<class Type = double>
class Matrix {
public:
	Matrix(unsigned rows, unsigned columns);
	Matrix(const Matrix<Type>& other);

	Matrix<Type>& operator=(const Matrix<Type>& other);

	void copy(const Matrix<Type>& other);

	bool operator==(const Matrix<Type>& other) const;
	bool operator!=(const Matrix<Type>& other) const;

	inline Type& operator()(unsigned row, unsigned column) { return matrix[column][row]; }
	inline Type operator()(unsigned row, unsigned column) const { return matrix[column][row]; }

	Matrix<Type>* operator*(const Matrix<Type>& other) const;

	inline unsigned getColumns() const { return columns; }
	inline unsigned getRows() const {  return rows; }

	inline const Type* getColumn(unsigned column) const { return matrix[column]; }
	void insertColumn(const Type* newColumn);

	std::string toString() const;

	virtual ~Matrix();
protected:
	void allocate();
	void deallocate();

	unsigned rows, columns;

	//internally represented as Matrix[column][row].
	Type** matrix;
};

template<class Type>
Matrix<Type>::Matrix(unsigned rows, unsigned columns)
	: rows(rows), columns(columns)
{
	if(columns == 0 || rows == 0)
		throw InvalidParameterException("matrix is empty.");

	allocate();
}

template<class Type>
Matrix<Type>::Matrix(const Matrix<Type>& other)
	: rows(other.rows), columns(other.columns)
{
	allocate();
	copy(other);
}

template<class Type>
Matrix<Type>& Matrix<Type>::operator=(const Matrix<Type>& other){
	if(this != &other){
		deallocate();
		rows = other.rows;
		columns = other.columns;
		allocate();
		copy(other);
	}

	return *this;
}

template<class Type>
void Matrix<Type>::copy(const Matrix<Type>& other){
	if(rows != other.rows || columns != other.columns)
		throw InvalidParameterException("matrix innapropriate to copy");

	for(unsigned i = 0; i < columns; i++)
		for(unsigned j = 0; j < rows; j++)
			matrix[i][j] = other.matrix[i][j];
}

template<class Type>
void Matrix<Type>::allocate(){
	matrix = new Type*[columns];

	for(unsigned i = 0; i < columns; i++)
		matrix[i] = new Type[rows];
}

template<class Type>
void Matrix<Type>::deallocate(){
	for(unsigned i = 0; i < columns; i++)
		delete [] matrix[i];

	delete [] matrix;
}

template<class Type>
Matrix<Type>* Matrix<Type>::operator*(const Matrix<Type>& other) const{
	if(columns != other.rows)
		throw InvalidParameterException("matrix innapropriate for multiplication");

	Matrix<Type> *matrixResult = new Matrix<Type>(rows,other.columns);

	for(unsigned i = 0; i < (*matrixResult).columns; i++)
		for(unsigned j = 0; j < (*matrixResult).rows; j++)
			(*matrixResult).matrix[i][j] = 0;

	for(unsigned i = 0; i < other.columns; i++)
		for(unsigned j = 0; j < rows; j++)
			for(unsigned k = 0; k < columns; k++)
				(*matrixResult).matrix[i][j] = (*matrixResult).matrix[i][j] + ( matrix[k][j] * other.matrix[i][k] );

	return matrixResult;
}

template<class Type>
bool Matrix<Type>::operator==(const Matrix<Type>& other) const{
	if(rows != other.rows || columns != other.columns)
		return false;

	for(unsigned i = 0; i < columns; i++)
		for(unsigned j = 0; j < rows; j++)
			if(matrix[i][j] != other.matrix[i][j])
				return false;

	return true;
}

template<class Type>
bool Matrix<Type>::operator!=(const Matrix<Type>& other) const {
	return !(*this == other);
}

template<class Type>
void Matrix<Type>::insertColumn(const Type* newColumn){
	if(!newColumn)
		throw NullPointerException("new column is null");

	Type **newMatrix = new Type*[columns + 1];

	for(unsigned i = 0; i < columns; i++)
		newMatrix[i] = matrix[i];

	newMatrix[columns] = new Type[rows];
	for(unsigned i = 0; i < rows; i++)
		newMatrix[columns][i] = newColumn[i];

	delete [] matrix;

	matrix = newMatrix;
	columns++;
}

template<class Type>
std::string Matrix<Type>::toString() const{
	std::stringstream strBuffer;

	for(unsigned i = 0; i < rows; i++){
		strBuffer << "| ";
		for(unsigned j = 0; j < columns; j++){
			strBuffer << matrix[j][i] << " ";
		}
		strBuffer << "|" << std::endl;
	}

	return strBuffer.str();
}

template<class Type>
Matrix<Type>::~Matrix(){
	deallocate();
}

#endif /* MATRIX_H_ */
