#include "BSpline.h"
#include "MatrixFactory.h"
#include "Polygon.h"

BSpline::BSpline(CoordinateList& points, const std::string name, unsigned subdivisions, const DrawCfg& drawCfg):
	Polyline(points, name, drawCfg), subdivisions(subdivisions){
	if(points.size() < 4)
		throw InvalidParameterException("invalid number of points to define a b-spline curve (min. 4)");

	type = GeometricElement::BSPLINE;
}

BSpline::BSpline(const BSpline& other) : Polyline(other), subdivisions(other.subdivisions){
	if(nPoints() < 4)
		throw InvalidParameterException("invalid number of points to define a b-spline curve (min. 4)");

	type = other.type;
}

GeometricElement* BSpline::clone() const{
	return new BSpline(*this);
}

LineList BSpline::getLines() const{
	LineList lines;

	CoordinateList coordinates = createCoordinates();
	for(unsigned i = 0; i < coordinates.size() - 1; i++){
		std::stringstream buffer;
		buffer << "#" << i;
		lines.push_back(Line(coordinates[i], coordinates[i+1], getName().append(buffer.str()), getDrawingConfig()));
	}

	return lines;
}

CoordinateList BSpline::createCoordinates() const{
	CoordinateList points(nPoints());
	for(unsigned i = 0; i < nPoints(); i++)
		points[i] = Coordinate((*matrix)(0,i),(*matrix)(1,i),(*matrix)(2,i));

	Matrix<double> geometryMatrix(4,3);
	double f0[3], deltaF0[3], delta2F0[3], delta3F0[3];

	CoordinateList curvePoints;

	Matrix<double> *BsplineMatrix = MatrixFactory::bSplineMatrix();
	for(unsigned iCurve = 0; iCurve < points.size() - 3 ; iCurve++){

		for(unsigned i = 0; i < geometryMatrix.getRows(); i++){
			geometryMatrix(i,0) = points[i + iCurve].x;
			geometryMatrix(i,1) = points[i + iCurve].y;
			geometryMatrix(i,2) = points[i + iCurve].z;
		}

		Matrix<double> *coeficients = (*BsplineMatrix) * geometryMatrix;

		double delta = 1.0/subdivisions;

		for(unsigned j = 0; j < 3; j++){
			f0[j] 		= (*coeficients)(3,j);
			deltaF0[j]  = (*coeficients)(0,j)*(delta*delta*delta) + (*coeficients)(1,j)*(delta*delta) + (*coeficients)(2,j)*delta;
			delta2F0[j] = 6*(*coeficients)(0,j)*(delta*delta*delta) + 2*(*coeficients)(1,j)*(delta*delta);
			delta3F0[j] = 6*(*coeficients)(0,j)*(delta*delta*delta);
		}

		delete coeficients;

		CoordinateList parcialPoints = forwardDiferences(points[iCurve],f0, deltaF0, delta2F0, delta3F0);
		for(unsigned j = 0; j < parcialPoints.size(); j++)
			curvePoints.push_back(parcialPoints[j]);
	}
	delete BsplineMatrix;

	return curvePoints;

}

CoordinateList BSpline::forwardDiferences(Coordinate& point, double f0[],
		double deltaf0[], double delta2f0[], double delta3f0[]) const
{
	CoordinateList curvePoints;

	for(unsigned i = 0; i < subdivisions; i++){
		point.x = f0[0];
		point.y = f0[1];
		point.z = f0[2];

		curvePoints.push_back(point);

		f0[0] += deltaf0[0]; deltaf0[0] += delta2f0[0]; delta2f0[0] += delta3f0[0];
		f0[1] += deltaf0[1]; deltaf0[1] += delta2f0[1]; delta2f0[1] += delta3f0[1];
		f0[2] += deltaf0[2]; deltaf0[2] += delta2f0[2]; delta3f0[2] += delta3f0[1];

	}

	return curvePoints;
}

TiXmlElement* BSpline::toXmlElement() const{
	TiXmlElement *polylineElement = new TiXmlElement("curvaBSpline");

	polylineElement->SetAttribute("nome",name.c_str());
	polylineElement->SetAttribute("cor", drawingConfig.getColor().colorCode().c_str());

	for(unsigned i = 0; i < nPoints(); i++){
		TiXmlElement *point = new TiXmlElement("ponto");
		point->SetDoubleAttribute("x",(*matrix)(0,i));
		point->SetDoubleAttribute("y",(*matrix)(1,i));
		point->SetDoubleAttribute("z",(*matrix)(2,i));

		polylineElement->LinkEndChild(point);
	}

	return polylineElement;
}

BSpline* BSpline::createFromXml(TiXmlElement* element){
	if(!element)
		throw InvalidParameterException("invalid XML.");

	CoordinateList coordinates;
	for(TiXmlElement *i = element->FirstChildElement("ponto"); i != 0; i = i->NextSiblingElement("ponto")){
		Coordinate coordinate;
		i->QueryDoubleAttribute("x",&coordinate.x);
		i->QueryDoubleAttribute("y",&coordinate.y);
		i->QueryDoubleAttribute("z",&coordinate.z);

		coordinates.push_back(coordinate);
	}

	std::string name(element->Attribute("nome"));
	DrawCfg drawingCfg(Color(std::string("#").append(element->Attribute("cor"))));

	return new BSpline(coordinates,name, 100, drawingCfg);
}

BSpline::~BSpline() {
}
