#include "Bezier.h"

Bezier::Bezier(CoordinateList& points, const std::string& name, unsigned subdivisions, const DrawCfg& drawCfg)
	: Polyline(points, name, drawCfg), subdivisions(subdivisions)
{
	if((nPoints() - 4) % 3 != 0)
		throw InvalidParameterException("invalid number of points to define a bezier curve.");

	type = GeometricElement::BEZIER;
}

Bezier::Bezier(const Bezier& other) : Polyline(other), subdivisions(other.subdivisions){
	if((nPoints() - 4) % 3 != 0)
		throw InvalidParameterException("invalid number of points to define a bezier curve.");

	type = other.type;
}

GeometricElement* Bezier::clone() const{
	return new Bezier(*this);
}

LineList Bezier::getLines() const{
	LineList lines;

	CoordinateList coordinates = createCoordinates();
	for(unsigned i = 0; i < coordinates.size() - 1; i++){
		std::stringstream buffer;
		buffer << "#" << i;
		lines.push_back(Line(coordinates[i], coordinates[i+1], getName().append(buffer.str()), getDrawingConfig()));
	}

	return lines;
}

CoordinateList Bezier::createCoordinates() const{
	CoordinateList points(nPoints());
	for(unsigned i = 0; i < nPoints(); i++)
		points[i] = Coordinate((*matrix)(0,i),(*matrix)(1,i),(*matrix)(2,i));

	unsigned curves = points.size() / 3;

	CoordinateList coordinates( (curves*subdivisions) + 1 );

	for(unsigned i = 0; i < curves; i++){
		CoordinateList iCurvePoints;
		for(unsigned j = 3 * i; j < 3 * i + 4; j++)
			iCurvePoints.push_back(Coordinate(points[j].x, points[j].y, points[j].z));

		double iterationPass = 0;
		unsigned endPoint = (i + 1) * subdivisions;

		for(unsigned j = i * subdivisions; j <= endPoint ; j++){
			coordinates[j] = blendingFunction(iterationPass, iCurvePoints);
			iterationPass += 1.0/(subdivisions);
		}
	}

	return coordinates;
}

Coordinate Bezier::blendingFunction(double t, const CoordinateList& points) const{
	double t2 = t*t;
	double t3 = t2 * t;

	Coordinate parameters;

	parameters.x = (points[0].x) * (-t3 + 3*t2 - 3*t + 1) + (points[1].x) * (3*t3 - 6*t2 + 3*t)	+
			(points[2].x) * (-3*t3 + 3*t2 ) + (points[3].x) * (t3);

	parameters.y = (points[0].y) * (-t3 + 3*t2 - 3*t + 1) + (points[1].y) * (3*t3 - 6*t2 + 3*t)	+
			 (points[2].y) * (-3*t3 + 3*t2 ) + (points[3].y) * (t3);

	parameters.z =  (points[0].z) * (-t3 + 3*t2 - 3*t + 1) + (points[1].z) * (3*t3 - 6*t2 + 3*t)	+
			 (points[2].z) * (-3*t3 + 3*t2 ) + (points[3].z) * (t3);

	return parameters;
}

TiXmlElement* Bezier::toXmlElement() const{
	TiXmlElement *bezierElement = new TiXmlElement("curvaBezier");

	bezierElement->SetAttribute("nome",name.c_str());
	bezierElement->SetAttribute("cor", drawingConfig.getColor().colorCode().c_str());

	unsigned nGeometries = (nPoints()-4)/3 + 1;
	for(unsigned i = 0; i < nGeometries; i++){
		TiXmlElement *geometryElement = new TiXmlElement("geometria");

		for(unsigned j = 0; j < 4; j++){
			std::stringstream buffer; buffer << "p" << j+1;
			TiXmlElement *point = new TiXmlElement(buffer.str().c_str());

			point->SetDoubleAttribute("x",(*matrix)(0,j + (i*3)));
			point->SetDoubleAttribute("y",(*matrix)(1,j + (i*3)));
			point->SetDoubleAttribute("z",(*matrix)(2,j + (i*3)));

			geometryElement->LinkEndChild(point);
		}

		bezierElement->LinkEndChild(geometryElement);
	}

	return bezierElement;
}

Bezier* Bezier::createFromXml(TiXmlElement* element){
	if(!element)
		throw InvalidParameterException("invalid XML.");

	bool firstGeometry = true;
	CoordinateList controlCoordinates;
	for(TiXmlElement *i = element->FirstChildElement("geometria"); i != 0; i = i->NextSiblingElement("geometria")){
		TiXmlElement *p[4];

		p[0] = i->FirstChildElement("p1");
		p[1] = i->FirstChildElement("p2");
		p[2] = i->FirstChildElement("p3");
		p[3] = i->FirstChildElement("p4");

		if(!(p[0] || p[1] || p[2] || p[3]))
			throw InvalidParameterException("invalid XML.");

		for(unsigned i = 0; i < 4; i++){
			if(!firstGeometry && i == 0)
				continue;

			Coordinate coordinate;
			p[i]->QueryDoubleAttribute("x",&coordinate.x);
			p[i]->QueryDoubleAttribute("y",&coordinate.y);
			p[i]->QueryDoubleAttribute("z",&coordinate.z);

			controlCoordinates.push_back(coordinate);
		}

		firstGeometry = false;
	}

	std::string name(element->Attribute("nome"));
	DrawCfg drawingCfg(Color(std::string("#").append(element->Attribute("cor"))));

	return new Bezier(controlCoordinates,name, 100, drawingCfg);
}

Bezier::~Bezier() {
}
