#include "BicubicSurface.h"
#include "MatrixFactory.h"

BicubicSurface::BicubicSurface(CoordinateList &points, const std::string& name, unsigned curveSubdivisions, unsigned nCurves, const DrawCfg& drawCfg)
	: Polyline(points, name, drawCfg), curveSubdivisions(curveSubdivisions), nCurves(nCurves) {
	if(nPoints() % 16 != 0)
		throw InvalidParameterException("invalid number of points to define a surface.");

	type = GeometricElement::SURFACE;
}

BicubicSurface::BicubicSurface(const BicubicSurface& other)
	: Polyline(other), curveSubdivisions(other.curveSubdivisions), nCurves(other.nCurves) {
	type = other.type;
}

LineList BicubicSurface::getLines() const{
	LineList lines;
	CoordinateList coordsS;
	CoordinateList coordsT;

	for(unsigned i = 0; i < nPoints()/16; i++){
		for(double j = 0; j <= 1 + (1.0/nCurves); j+= 1.0/nCurves ){
			coordsS = createCurveS(j,i);
			coordsT = createCurveT(j,i);

			for(unsigned k = 0; k < coordsS.size() - 1; k++){
				std::stringstream buffer;
				buffer << "#" << j << "#" << k << "#" << i;
				lines.push_back(Line(coordsS[k], coordsS[k+1], getName().append(buffer.str()).append("#1"),getDrawingConfig()));
				lines.push_back(Line(coordsT[k], coordsT[k+1], getName().append(buffer.str()).append("#2"),getDrawingConfig()));
			}
		}
	}

	return lines;
}

CoordinateList BicubicSurface::createCurveS(double s, unsigned geoNumber) const{
	Matrix<Coordinate>* geoMatrix = MatrixFactory::surfaceGeoMatrix(matrix, geoNumber);
	Matrix<double> xMatrix(4,4), yMatrix(4,4), zMatrix(4,4);

	for(unsigned i = 0; i < (*geoMatrix).getRows(); i++)
		for(unsigned j = 0; j < (*geoMatrix).getColumns(); j++){
			xMatrix(i,j) = (*geoMatrix)(i,j).x;
			yMatrix(i,j) = (*geoMatrix)(i,j).y;
			zMatrix(i,j) = (*geoMatrix)(i,j).z;
		}
	Matrix<double> *matrixes[] = {&xMatrix, &yMatrix, &zMatrix};

	double pass = 1.0/curveSubdivisions;
	CoordinateList points;

	for(double j = 0; j <= 1.01; j += pass){
		points.push_back(blend(s,j,matrixes));
	}

	delete geoMatrix;

	return points;
}

CoordinateList BicubicSurface::createCurveT(double t, unsigned geoNumber) const{
	Matrix<Coordinate>* geoMatrix = MatrixFactory::surfaceGeoMatrix(matrix, geoNumber);
	Matrix<double> xMatrix(4,4), yMatrix(4,4), zMatrix(4,4);

	for(unsigned i = 0; i < (*geoMatrix).getRows(); i++)
		for(unsigned j = 0; j < (*geoMatrix).getColumns(); j++){
			xMatrix(i,j) = (*geoMatrix)(i,j).x;
			yMatrix(i,j) = (*geoMatrix)(i,j).y;
			zMatrix(i,j) = (*geoMatrix)(i,j).z;
		}
	Matrix<double> *matrixes[] = {&xMatrix, &yMatrix, &zMatrix};

	double pass = 1.0/curveSubdivisions;
	CoordinateList points;

	for(double j = 0; j <= 1.01; j += pass){
		points.push_back(blend(j,t,matrixes));
	}

	delete geoMatrix;

	return points;
}

Coordinate BicubicSurface::blend(double s, double t, Matrix<double>* matrixes[]) const{
	double s2 = s*s;
	double s3 = s2*s;
	double t2 = t*t;
	double t3 = t2*t;
	Matrix<double> S(1,4), T(4,1);
	S(0,0) = s3;	S(0,1) = s2;	S(0,2) = s; 	S(0,3) = 1;
	T(0,0) = t3;	T(1,0) = t2;	T(2,0) = t; 	T(3,0) = 1;

	Matrix<double> *bezierM = MatrixFactory::bezierMatrix();
	double coords[3];

	for(unsigned i = 0; i < 3; i++){
		Matrix<double>* partial, *tmp;
		partial = S*(*bezierM);
		tmp = partial;

		partial = (*partial)*(*matrixes[i]);
		delete tmp; tmp = partial;

		partial = (*partial)*(*bezierM);
		delete tmp; tmp = partial;

		partial = (*partial)*T;
		delete tmp;

		coords[i] = (*partial)(0,0);
		delete partial;
	}

	Coordinate blended(coords[0], coords[1], coords[2]);

	delete bezierM;

	return blended;
}

GeometricElement* BicubicSurface::clone() const{
	return new BicubicSurface(*this);
}

TiXmlElement* BicubicSurface::toXmlElement() const{
	TiXmlElement *surfaceElement = new TiXmlElement("superficieBezier");

	surfaceElement->SetAttribute("nome",name.c_str());
	surfaceElement->SetAttribute("cor", drawingConfig.getColor().colorCode().c_str());

	unsigned nGeometries = nPoints()/16;
	unsigned geometryStartsAt = 0;
	for(unsigned i = 0; i < nGeometries; i++){
		TiXmlElement *geometryElement = new TiXmlElement("geometria");

		for(unsigned j = 0; j < 16; j++){
			TiXmlElement *point = new TiXmlElement("elemento");


			point->SetDoubleAttribute("x",(*matrix)(0,j + geometryStartsAt));
			point->SetDoubleAttribute("y",(*matrix)(1,j + geometryStartsAt));
			point->SetDoubleAttribute("z",(*matrix)(2,j + geometryStartsAt));

			geometryElement->LinkEndChild(point);
		}

		geometryStartsAt = geometryStartsAt + 16;

		surfaceElement->LinkEndChild(geometryElement);
	}

	return surfaceElement;
}

BicubicSurface* BicubicSurface::createFromXml(TiXmlElement* element){
	if(!element)
		throw InvalidParameterException("invalid XML.");

	CoordinateList coordinates;
	for(TiXmlElement *i = element->FirstChildElement("geometria"); i != 0; i = i->NextSiblingElement("geometria")){
		for(TiXmlElement *j = i->FirstChildElement("elemento"); j != 0; j = j->NextSiblingElement("elemento")){
			Coordinate coordinate;
			j->QueryDoubleAttribute("x",&coordinate.x);
			j->QueryDoubleAttribute("y",&coordinate.y);
			j->QueryDoubleAttribute("z",&coordinate.z);

			coordinates.push_back(coordinate);
		}
	}

	std::string name(element->Attribute("nome"));
	DrawCfg drawingCfg(Color(std::string("#").append(element->Attribute("cor"))));

	return new BicubicSurface(coordinates, name, 20, 20, drawingCfg);
}

BicubicSurface::~BicubicSurface() {
}
