#include "Display.h"
#include "MatrixFactory.h"
#include "Element3D.h"
#include "BSpline.h"
#include "Bezier.h"
#include "BicubicSurface.h"

#include <iostream>
Display::Display(const Window& window, const Coordinate& planeNormal, const Coordinate& viewUp)
	: planeNormal(Coordinate(0,0,-1),"#planeNormal"),viewUp(Coordinate(0,1,0),"#viewUp"), window(window) {
	wOrientation = MatrixFactory::identityMatrix(4);
	iwOrientation = MatrixFactory::identityMatrix(4);

	rotateWindow(MatrixFactory::Y, -planeNormal.angleXZ(this->planeNormal.getCoordinate()));
	rotateWindow(MatrixFactory::X, planeNormal.angleYZ(this->planeNormal.getCoordinate()));

//	fixme: wrong
//	rotateWindow(MatrixFactory::Z, viewUp.angleXY(this->viewUp.getCoordinate()));
}

DisplayFile::Iterator Display::begin(){
	return displayFile.begin();
}

DisplayFile::Iterator Display::end(){
	return displayFile.end();
}

void Display::drawOn(Viewport& viewport){
	window.draw(ppcFile,viewport);
}

void Display::add(GeometricElement* element){
	displayFile.add(element);

	GeometricElement* ppcElement = element->clone();
	ppcElement->transform(*wOrientation);
	ppcFile.add(ppcElement);
}

const GeometricElement* Display::getElement(const std::string& name){
	return displayFile[name];
}

//elements
void Display::translate(const std::string& elementName, double xOffset, double yOffset){
	ppcFile[elementName]->translate(xOffset,yOffset,0);

	GeometricElement *element = displayFile[elementName];
	element->transform(*iwOrientation);
	element->translate(xOffset,yOffset,0);
	element->transform(*wOrientation);
}

void Display::scale(const std::string& elementName, double xScale, double yScale, double zScale){
	ppcFile[elementName]->scale(xScale,yScale, zScale);
	displayFile[elementName]->scale(xScale,yScale, zScale);
}

void Display::rotate(const std::string& elementName, MatrixFactory::Axis axis, double angle, const Coordinate& reference){
	displayFile[elementName]->rotate(axis, angle, reference);

	Point point(reference,"");
	point.transform(*wOrientation);
	ppcFile[elementName]->rotate(axis, angle, point.getCoordinate());
}

void Display::rotate(const std::string& elementName, MatrixFactory::Axis axis, double angle){
	ppcFile[elementName]->rotate(axis, angle);
	displayFile[elementName]->rotate(axis, angle);
}

//window
void Display::translateWindow(double xOffset, double yOffset, double zOffset){
	window.translate(xOffset,yOffset,zOffset);
}

void Display::scaleWindow(double xRatio, double yRatio, double zRatio){
	window.scale(xRatio,yRatio, zRatio);
}

void Display::rotateWindow(MatrixFactory::Axis axis, double angle){
	Matrix<double>* rotationMatrix = MatrixFactory::rotationMatrix(axis, angle, window.getCenter());

	for(DisplayFile::Iterator i = ppcFile.begin(); i != ppcFile.end(); i++)
		(*i)->transform(*rotationMatrix);

	delete rotationMatrix;

	rotateWindowOrientation(axis, angle);
}

void Display::rotateWindowOrientation(MatrixFactory::Axis axis, double angle){
	Matrix<double>* oldOrientation = wOrientation;
	Matrix<double>* rotationMatrix = MatrixFactory::rotationMatrix(axis, angle, window.getCenter());
	wOrientation = (*rotationMatrix) * (*wOrientation);

	delete oldOrientation;
	delete rotationMatrix;

	Matrix<double>* oldIwOrientation = iwOrientation;
	Matrix<double>* iRotationMatrix = MatrixFactory::rotationMatrix(axis, -angle, window.getCenter());
	iwOrientation = (*iwOrientation) * (*iRotationMatrix);

	delete oldIwOrientation;
	delete iRotationMatrix;

	planeNormal.rotate(axis,-angle,window.getCenter());
	viewUp.rotate(axis,-angle, window.getCenter());

//	std::cout << this->planeNormal.toString() << this->viewUp.toString() << std::endl;
}

Display::~Display() {
	delete wOrientation;
	delete iwOrientation;
}

void Display::save(const std::string& filepath){
	TiXmlDocument document;
	TiXmlDeclaration *version = new TiXmlDeclaration("1.0","","");
	document.LinkEndChild(version);

	TiXmlElement *formasElement = new TiXmlElement("formas");

	TiXmlElement* windowElement = new TiXmlElement("window");
	const char* wAttributes[] = { "cop", "vpn", "vup", "wmin", "wmax" };
	CoordinateList wParameters(5);
	wParameters[0] = window.getProjectionCenter();
	wParameters[1] = planeNormal.getCoordinate();
	wParameters[2] = viewUp.getCoordinate();
	wParameters[3] = window.min();
	wParameters[4] = window.max();

	for(unsigned i = 0; i < 5; i++){
		TiXmlElement* element = new TiXmlElement(wAttributes[i]);
		element->SetDoubleAttribute("x",wParameters[i].x);
		element->SetDoubleAttribute("y",wParameters[i].y);
		element->SetDoubleAttribute("z",wParameters[i].z);

		windowElement->LinkEndChild(element);
	}
	formasElement->LinkEndChild(windowElement);

	for(DisplayFile::Iterator i = begin(); i != end(); i++){
		TiXmlElement *xmlElement = (*i)->toXmlElement();
		formasElement->LinkEndChild(xmlElement);
	}

	document.LinkEndChild(formasElement);

	if(!document.SaveFile(filepath.c_str()))
		throw InvalidParameterException(std::string("invalid destination file: ").append(filepath));
}

Display* Display::load(const std::string& filepath){
	TiXmlDocument document( filepath.c_str() );
	if(!document.LoadFile())
		throw InvalidParameterException(std::string("invalid input file: ").append(filepath));

	TiXmlHandle hDoc(&document);
	TiXmlHandle formasHandle = hDoc.FirstChild("formas");

	TiXmlElement *windowElement = formasHandle.FirstChild("window").Element();
	if(!windowElement)
		throw InvalidParameterException("window not found in XML");

	//stores cop, vpn, vup, wmin, wmax, in this order.
	CoordinateList wParameters(5);
	const char* wAttributes[] = { "cop", "vpn", "vup", "wmin", "wmax" };

	for(unsigned i = 0; i < 5; i++){
		TiXmlElement *element = windowElement->FirstChildElement(wAttributes[i]);
		if(!element)
			throw InvalidParameterException(std::string(wAttributes[i]).append(" not found in window XML."));
		element->QueryDoubleAttribute("x",&wParameters[i].x);
		element->QueryDoubleAttribute("y",&wParameters[i].y);
		element->QueryDoubleAttribute("z",&wParameters[i].z);
	}

	Display *display = new Display(::Window(wParameters[3],wParameters[4],wParameters[0]),wParameters[1], wParameters[2]);

	for(TiXmlElement *points = formasHandle.FirstChild("ponto").Element(); points != 0; points = points->NextSiblingElement("ponto"))
		display->add(Point::createFromXml(points));

	for(TiXmlElement *lines = formasHandle.FirstChild("reta").Element(); lines != 0; lines = lines->NextSiblingElement("reta"))
		display->add(Line::createFromXml(lines));

	for(TiXmlElement *polylines = formasHandle.FirstChild("polilinha").Element(); polylines != 0; polylines = polylines->NextSiblingElement("polilinha"))
		display->add(Polyline::createFromXml(polylines));

	for(TiXmlElement *polygons = formasHandle.FirstChild("poligono").Element(); polygons != 0; polygons = polygons->NextSiblingElement("poligono"))
		display->add(Polygon::createFromXml(polygons));

	for(TiXmlElement *elements3d = formasHandle.FirstChild("objeto3d").Element(); elements3d != 0; elements3d = elements3d->NextSiblingElement("objeto3d"))
		display->add(Element3D::createFromXml(elements3d));

	for(TiXmlElement *bsplines = formasHandle.FirstChild("curvaBSpline").Element(); bsplines != 0; bsplines = bsplines->NextSiblingElement("curvaBSpline"))
		display->add(BSpline::createFromXml(bsplines));

	for(TiXmlElement *beziers = formasHandle.FirstChild("curvaBezier").Element(); beziers != 0; beziers = beziers->NextSiblingElement("curvaBezier"))
		display->add(Bezier::createFromXml(beziers));

	for(TiXmlElement *surface = formasHandle.FirstChild("superficieBezier").Element(); surface != 0; surface = surface->NextSiblingElement("superficieBezier"))
		display->add(BicubicSurface::createFromXml(surface));

	return display;
}
