#include "Element3D.h"

Element3D::Element3D(const CoordinateList& coordinates, const EdgeList& edges, const std::string& name,
		const DrawCfg& drawCfg)
	: Polyline(coordinates, name, drawCfg), edges(edges)
	{
	for(unsigned i = 0; i < edges.size(); i++)
		if(edges[i].first > nPoints() || edges[i].second > nPoints())
			throw InvalidParameterException("edges don't connect valid coordinates");

	type = GeometricElement::ELEMENT3D;
}

Element3D::Element3D(const Element3D& other) : Polyline(other), edges(other.edges){
	type = GeometricElement::ELEMENT3D;
}

GeometricElement* Element3D::clone() const{
	return new Element3D(*this);
}

LineList Element3D::getLines() const{
	LineList lines;

	for(unsigned i = 0; i < edges.size(); i++){
		Coordinate coord1(matrix->getColumn(edges[i].first));
		Coordinate coord2(matrix->getColumn(edges[i].second));

		std::stringstream buffer;
		buffer << "#" << i;
		lines.push_back(Line(coord1, coord2, getName().append(buffer.str()), getDrawingConfig()));
	}

	return lines;
}

std::string Element3D::toString() const{
	std::stringstream buffer;
	buffer << GeometricElement::toString();

	for(unsigned i = 0; i < edges.size(); i++)
		buffer << "( " << edges[i].first << ", " << edges[i].second << ")" << std::endl;

	return buffer.str();
}

TiXmlElement* Element3D::toXmlElement() const{
	TiXmlElement *xml3dElement = new TiXmlElement("objeto3d");

	xml3dElement->SetAttribute("nome",name.c_str());
	xml3dElement->SetAttribute("cor", drawingConfig.getColor().colorCode().c_str());

	TiXmlElement *pontosElement = new TiXmlElement("pontos");
	for(unsigned i = 0; i < nPoints(); i++){
		TiXmlElement *point = new TiXmlElement("ponto");
		point->SetDoubleAttribute("x",(*matrix)(0,i));
		point->SetDoubleAttribute("y",(*matrix)(1,i));
		point->SetDoubleAttribute("z",(*matrix)(2,i));

		pontosElement->LinkEndChild(point);
	}
	xml3dElement->LinkEndChild(pontosElement);

	TiXmlElement *arestasElement = new TiXmlElement("arestas");
	for(unsigned i = 0; i < edges.size(); i++){
		TiXmlElement *aresta = new TiXmlElement("aresta");
		aresta->SetAttribute("v1",edges[i].first);
		aresta->SetAttribute("v2",edges[i].second);

		arestasElement->LinkEndChild(aresta);
	}

	xml3dElement->LinkEndChild(arestasElement);

	return xml3dElement;
}

Element3D* Element3D::createFromXml(TiXmlElement* element){
	if(!element)
		throw InvalidParameterException("invalid XML.");

	TiXmlElement *pontosElement = element->FirstChildElement("pontos");
	if(!pontosElement)
		throw InvalidParameterException("invalid XML.");

	TiXmlElement *arestasElement = element->FirstChildElement("arestas");
	if(!arestasElement)
		throw InvalidParameterException("invalid XML.");

	CoordinateList coordinates;
	for(TiXmlElement *i = pontosElement->FirstChildElement("ponto"); i != 0; i = i->NextSiblingElement("ponto")){
		Coordinate coordinate;
		i->QueryDoubleAttribute("x",&coordinate.x);
		i->QueryDoubleAttribute("y",&coordinate.y);
		i->QueryDoubleAttribute("z",&coordinate.z);

		coordinates.push_back(coordinate);
	}

	EdgeList edges;
	for(TiXmlElement *i = arestasElement->FirstChildElement("aresta"); i != 0; i = i->NextSiblingElement("aresta")){
		int v1,v2;
		i->QueryIntAttribute("v1",&v1);
		i->QueryIntAttribute("v2",&v2);

		edges.push_back(std::pair<unsigned, unsigned>((unsigned)v1,(unsigned)v2));
	}

	std::string name(element->Attribute("nome"));
	DrawCfg drawingCfg(Color(std::string("#").append(element->Attribute("cor"))));

	return new Element3D(coordinates,edges,name,drawingCfg);
}

Element3D::~Element3D() {
}
