#include "GeometricElement.h"
#include "MatrixFactory.h"
#include "exception/GSException.h"
#include <cmath>

#include "Point.h"
#include "Polyline.h"
#include "Polygon.h"
#include "Bezier.h"
#include "BSpline.h"
#include "Element3D.h"
#include "BicubicSurface.h"

const std::string GeometricElement::ELEMENT_TYPES_STR[] = {
		"Unspecified", "Point", "Line", "Polyline", "Polygon", "Polygon", "Bezier", "BSPline", "3D", "Surface"
};

GeometricElement::GeometricElement(unsigned nPoints, GeometricElement::ElementType type, const std::string& name, const DrawCfg& drawingConfig)
	: center(0,0,0), name(name), type(type), drawingConfig(drawingConfig)
{
	if(nPoints == 0)
		throw InvalidParameterException("geometric element without points.");

	matrix = new Matrix<double>(4,nPoints);
}

GeometricElement::GeometricElement(const GeometricElement& other)
	: center(other.center), name(other.name), type(other.type), drawingConfig(other.drawingConfig)
{
	matrix = new Matrix<double>(*(other.matrix));
}

bool GeometricElement::operator==(const GeometricElement& other) const{
	return (*matrix) == (*other.matrix) && type == other.type;
}

bool GeometricElement::operator!=(const GeometricElement& other) const{
	return ! ( (*this) == other );
}

void GeometricElement::translate(double xOffset, double yOffset, double zOffset){
	Matrix<double> *tMatrix = MatrixFactory::translationMatrix(xOffset, yOffset, zOffset);
	transform(*tMatrix);
	delete tMatrix;
}

void GeometricElement::scale(double xScale, double yScale, double zScale, const Coordinate& reference){
	Matrix<double> *tMatrix = MatrixFactory::scalingMatrix(xScale, yScale, zScale, reference);
	transform(*tMatrix);
	delete tMatrix;
}

void GeometricElement::rotate(MatrixFactory::Axis axis, double angle, const Coordinate& reference){
	Matrix<double> *tMatrix = MatrixFactory::rotationMatrix(axis, angle, reference);
	transform(*tMatrix);
	delete tMatrix;
}

void GeometricElement::scale(double xScale, double yScale, double zScale){
	scale(xScale, yScale, zScale, center);
}

void GeometricElement::rotate(MatrixFactory::Axis axis, double angle){
	rotate(axis, angle, center);
}

void GeometricElement::transform(const Matrix<double>& transformationMatrix){
	Matrix<double>* oldMatrix = matrix;
	matrix =  transformationMatrix * (*matrix) ;
	delete oldMatrix;

	recalculateCenter();
}

void GeometricElement::project(const Rectangle& planeRectangle, const Coordinate& projectionCenter){
	translate(-projectionCenter.x,-projectionCenter.y,-projectionCenter.z);

	//fixme: is this right?
	double d = (planeRectangle.getCenter().z - projectionCenter.z);

	for(unsigned i = 0; i < nPoints(); i++){
		double w = (*matrix)(2,i)/d;

		(*matrix)(0,i) = (*matrix)(0,i)/w;
		(*matrix)(1,i) = (*matrix)(1,i)/w;
		(*matrix)(2,i) = (*matrix)(2,i)/w;

		(*matrix)(3,i) = 1;
	}

	translate(projectionCenter.x,projectionCenter.y,projectionCenter.z);
}

void GeometricElement::recalculateCenter(){
	unsigned nPoints = matrix->getColumns();

	double sumX = 0, sumY = 0, sumZ = 0;
	for(unsigned i = 0; i < nPoints; i++){
		sumX += (*matrix)(0,i);
		sumY += (*matrix)(1,i);
		sumZ += (*matrix)(2,i);
	}

	center.x = sumX/nPoints ; center.y = sumY/nPoints ; center.z = sumZ/nPoints;
}

ElementsList GeometricElement::clip(const Rectangle& clippingArea) const{
	ElementsList elements;
	elements.push_back(this->clone());
	return elements;
}

ElementsList GeometricElement::clip3d(const Rectangle& clippingArea, const Coordinate& projectionCenter) const{
	ElementsList elements;
	elements.push_back(this->clone());
	return elements;
}

std::string GeometricElement::getName() const{
	return name;
}

std::string GeometricElement::toString() const{
	std::stringstream buffer;
	buffer << "name: " << name << ". type: " << GeometricElement::ELEMENT_TYPES_STR[type];
	buffer << ". drawing: " << drawingConfig.toString();
	buffer<< ". matrix: " << std::endl;
	buffer << matrix->toString() << std::endl;

	return buffer.str();
}

Coordinate GeometricElement::getCenter() const{
	return center;
}

DrawCfg GeometricElement::getDrawingConfig() const{
	return drawingConfig;
}

bool GeometricElement::is(ElementType type) const{
	return this->type == type;
}

GeometricElement::ElementType GeometricElement::getType() const{
	return type;
}

void GeometricElement::save(const std::string& filepath) const{
	TiXmlDocument document;
	TiXmlDeclaration *version = new TiXmlDeclaration("1.0","","");
	document.LinkEndChild(version);

	document.LinkEndChild(toXmlElement());

	if(!document.SaveFile(filepath.c_str()))
		throw InvalidParameterException(std::string("invalid destination file: ").append(filepath));
}

GeometricElement* GeometricElement::load(const std::string& filepath){
	TiXmlDocument document( filepath.c_str() );
	if(!document.LoadFile())
		throw InvalidParameterException(std::string("invalid input file: ").append(filepath));

	TiXmlHandle hDoc(&document);

	TiXmlElement *xmlElement = hDoc.FirstChild("ponto").Element();
	if(xmlElement)
		return Point::createFromXml(xmlElement);

	xmlElement = hDoc.FirstChild("reta").Element();
	if(xmlElement)
		return Line::createFromXml(xmlElement);

	xmlElement = hDoc.FirstChild("polilinha").Element();
	if(xmlElement)
		return Polyline::createFromXml(xmlElement);

	xmlElement = hDoc.FirstChild("poligono").Element();
	if(xmlElement)
		return Polygon::createFromXml(xmlElement);

	xmlElement = hDoc.FirstChild("objeto3d").Element();
	if(xmlElement)
		return Element3D::createFromXml(xmlElement);

	xmlElement = hDoc.FirstChild("curvaBSpline").Element();
	if(xmlElement)
		return BSpline::createFromXml(xmlElement);

	xmlElement = hDoc.FirstChild("curvaBezier").Element();
	if(xmlElement)
		return Bezier::createFromXml(xmlElement);

	xmlElement = hDoc.FirstChild("superficieBezier").Element();
	if(xmlElement)
		return BicubicSurface::createFromXml(xmlElement);

	throw InvalidParameterException("invalid object XML.");
}

GeometricElement::~GeometricElement() {
	delete matrix;
}
