#include "Line.h"
#include "Polygon.h"
#include <algorithm>

Line::Line(const Coordinate& point1, const Coordinate& point2, const std::string& name, const DrawCfg& drawCfg)
	: GeometricElement(2, GeometricElement::LINE, name, drawCfg){
	(*matrix)(0,0) = point1.x;
	(*matrix)(1,0) = point1.y;
	(*matrix)(2,0) = point1.z;
	(*matrix)(3,0) = 1;

	(*matrix)(0,1) = point2.x;
	(*matrix)(1,1) = point2.y;
	(*matrix)(2,1) = point2.z;
	(*matrix)(3,1) = 1;

	recalculateCenter();
}

GeometricElement* Line::clone() const{
	return new Line(*this);
}

Coordinate Line::first() const{
	return Coordinate((*matrix)(0,0), (*matrix)(1,0), (*matrix)(2,0));
}

Coordinate Line::second() const{
	return Coordinate((*matrix)(0,1), (*matrix)(1,1), (*matrix)(2,1));
}

bool Line::clipLineOnEdge(double denominator, double numerator, double& tE, double& tL) const{
	double t;
	bool accept = true;

	if(denominator > 0){
		t = numerator/denominator;

		if(t > tL)
			accept = false;
		else{
			if(t > tE)
				tE = t;
		}
	}
	else{
		if(denominator < 0){
			t = numerator/denominator;

			if(t < tE)
				accept = false;
			else {
				if(t < tL)
					tL = t;
			}
		}
		else{
			if(numerator > 0)
				accept = false;
		}
	}

	return accept;
}

// Liang-Barsky parametric line-clipping algorithm.
//Reference:  Foley, James D., Computer Graphics: Principles and pratices, pg 122.
ElementsList Line::clip(const Rectangle& clippingArea) const{
	ElementsList list;

	double x0 = (*matrix)(0,0); double x1 = (*matrix)(0,1);
	double y0 = (*matrix)(1,0); double y1 = (*matrix)(1,1);

	Coordinate wMax = clippingArea.max();
	Coordinate wMin = clippingArea.min();

	double xMin = wMin.x; double yMin = wMin.y;
	double xMax = wMax.x; double yMax = wMax.y;

	double deltaX = x1 - x0;
	double deltaY = y1 - y0;

	//line is just a point
	if(deltaX == 0 && deltaY == 0)
		return list;

	//scalars used to determine the new extreme points
	double tE = 0;
	double tL = 1;

	if( clipLineOnEdge(deltaX, xMin - x0, tE, tL) && clipLineOnEdge(-deltaX, x0 - xMax, tE, tL) &&
			clipLineOnEdge(deltaY, yMin - y0, tE, tL) && clipLineOnEdge(-deltaY, y0 - yMax, tE, tL) ) {

		if(tL < 1){
			x1 = x0 + (tL * deltaX);
			y1 = y0 + (tL * deltaY);
		}
		if(tE > 0){
			x0 = x0 + (tE * deltaX);
			y0 = y0 + (tE * deltaY);
		}

		list.push_back(new Line(Coordinate(x0,y0,(*matrix)(2,0)), Coordinate(x1,y1,(*matrix)(2,1)), name, drawingConfig));
	}

	return list;
}

//fixme: doesn't work
// Liang-Barsky parametric line-clipping algorithm 3d extension.
//Reference:  Foley, James D., Computer Graphics: Principles and pratices, pg 274.
ElementsList Line::clip3d(const Rectangle& clippingArea, const Coordinate& projectionCenter) const{
	ElementsList elements;
	elements.push_back(this->clone());
	return elements;
//
//	double x0 = first().x; double y0 = first().y; double z0 = first().z;
//	double x1 = second().x; double y1 = second().y; double z1 = second().z;
//
//	bool accept = false;
//
//	double tmin = 0;
//	double tmax = 1;
//	double dx = x1 - x0;
//	double dz = z1 - z0;
//
//	if(clipLineOnEdge(-dx -dz, x0 + z0, tmin,tmax)){  //right side
//		if(clipLineOnEdge(dx - dz, -x0 + z0, tmin, tmax)){ //left side
//			double dy = y1 - y0;
//
//			if(clipLineOnEdge(dy - dz, -y0 + z0, tmin, tmax)){ //bottom
//				if(clipLineOnEdge(-dy -dz, y0 + z0 ,tmin,tmax)){ //top
//					if(clipLineOnEdge(-dz,z0 -zmin, tmin, tmax)){ //front
//						if(clipLineOnEdge(dz,-z0 -1, tmin,tmax)){ //back
//							accept = true;
//
//							if(tmax < 1){
//								x1 = x0 + tmax * dx;
//								y1 = y0 + tmax * dy;
//								z1 = z0 + tmax * dz;
//							}
//
//							if(tmin > 0){
//								x0 = x0 + tmin * dx;
//								y0 = y0 + tmin * dy;
//								z0 = z0 + tmin * dz;
//							}
//						}
//					}
//				}
//			}
//		}
//	} // wow
//
//	ElementsList elements;
//	if(accept)
//		elements.push_back(new Line(Coordinate(x0,y0,z0),Coordinate(x1,y1,z1),name,drawingConfig));
//
//	return elements;
}

Coordinate* Line::intersects(const Line& other){
	Coordinate _first = first(); Coordinate _second = second();
	Coordinate oFirst = other.first(); Coordinate oSecond = other.second();

	double x2 = _second.x; double x1 = _first.x;
	double y2 = _second.y; double y1 = _first.y;
	double dx = x2 - x1;
	double dy = y2 - y1;

	double a2 = oSecond.x; double a1 = oFirst.x;
	double b2 = oSecond.y; double b1 = oFirst.y;
	double da = a2 - a1;
	double db = b2 - b1;

	//parallel
	if(da * dy == db * dx)
		return 0;

	double s = (dx * (b1 - y1) + dy * (x1 - a1))/ (da * dy - db* dx);
	double t = (da * (y1 - b1) + db * (a1 - x1)) / (db * dx - da*dy);

	if(s >= 0 && s <=1 && t >=0 && t <=1)
		return new Coordinate(x1 + t*dx, y1 + t*dy, (*matrix)(2,0) );

	return 0;
}

TiXmlElement* Line::toXmlElement() const{
	TiXmlElement *lineElement = new TiXmlElement("reta");

	lineElement->SetAttribute("nome",name.c_str());
	lineElement->SetAttribute("cor", drawingConfig.getColor().colorCode().c_str());

	for(unsigned i = 0; i < nPoints(); i++){
		TiXmlElement *point = new TiXmlElement("ponto");
		point->SetDoubleAttribute("x",(*matrix)(0,i));
		point->SetDoubleAttribute("y",(*matrix)(1,i));
		point->SetDoubleAttribute("z",(*matrix)(2,i));

		lineElement->LinkEndChild(point);
	}

	return lineElement;
}

Line* Line::createFromXml(TiXmlElement* element){
	if(!element)
		throw InvalidParameterException("invalid XML.");

	TiXmlElement *firstPoint = element->FirstChildElement("ponto");
	if(!firstPoint)
		throw InvalidParameterException("invalid XML.");
	TiXmlElement *secondPoint = firstPoint->NextSiblingElement("ponto");
	if(!secondPoint)
		throw InvalidParameterException("invalid XML.");

	Coordinate coord1, coord2;
	firstPoint->QueryDoubleAttribute("x",&coord1.x);
	firstPoint->QueryDoubleAttribute("y",&coord1.y);
	firstPoint->QueryDoubleAttribute("z",&coord1.z);

	secondPoint->QueryDoubleAttribute("x",&coord2.x);
	secondPoint->QueryDoubleAttribute("y",&coord2.y);
	secondPoint->QueryDoubleAttribute("z",&coord2.z);

	std::string name(element->Attribute("nome"));
	DrawCfg drawingCfg(Color(std::string("#").append(element->Attribute("cor"))));

	return new Line(coord1,coord2,name,drawingCfg);
}

Line::~Line() {
}
