#include "MatrixFactory.h"

Matrix<double>* MatrixFactory::identityMatrix(unsigned order){
	Matrix<double> *matrix = new Matrix<double>(order,order);

	for(unsigned i = 0; i < order; i++)
		for(unsigned j = 0; j < order; j++)
			(*matrix)(i,j) = 0;

	for(unsigned i = 0; i < order; i++)
		(*matrix)(i,i) = 1;


	return matrix;
}

//Reference:  Foley, James D., Computer Graphics: Principles and pratices, 2nd Ed, page 215.
Matrix<double>* MatrixFactory::scalingMatrix(double xScale, double yScale, double zScale, const Coordinate& reference){
	Matrix<double> *matrix = new Matrix<double>(4,4);
	(*matrix)(0,0) = xScale; 	(*matrix)(0,1) = 0; 		(*matrix)(0,2) = 0; 		(*matrix)(0,3) = -reference.x * (xScale -1);
	(*matrix)(1,0) = 0; 		(*matrix)(1,1) = yScale; 	(*matrix)(1,2) = 0; 		(*matrix)(1,3) = -reference.y * (yScale -1);
	(*matrix)(2,0) = 0; 		(*matrix)(2,1) = 0; 		(*matrix)(2,2) = zScale; 	(*matrix)(2,3) = -reference.z * (zScale -1);
	(*matrix)(3,0) = 0; 		(*matrix)(3,1) = 0; 		(*matrix)(3,2) = 0; 		(*matrix)(3,3) = 1;

	return matrix;
}

//Reference:  Foley, James D., Computer Graphics: Principles and pratices, 2nd Ed, page 214.
Matrix<double>* MatrixFactory::translationMatrix(double xOffset, double yOffset, double zOffset){
	Matrix<double> *matrix = new Matrix<double>(4,4);
	(*matrix)(0,0) = 1; (*matrix)(0,1) = 0; (*matrix)(0,2) = 0; (*matrix)(0,3) = xOffset;
	(*matrix)(1,0) = 0; (*matrix)(1,1) = 1; (*matrix)(1,2) = 0; (*matrix)(1,3) = yOffset;
	(*matrix)(2,0) = 0; (*matrix)(2,1) = 0; (*matrix)(2,2) = 1; (*matrix)(2,3) = zOffset;
	(*matrix)(3,0) = 0; (*matrix)(3,1) = 0; (*matrix)(3,2) = 0; (*matrix)(3,3) = 1;

	return matrix;
}

//Reference:  Foley, James D., Computer Graphics: Principles and pratices, 2nd Ed, page 215.
Matrix<double>* MatrixFactory::rotationMatrix(MatrixFactory::Axis axis, double angle, const Coordinate& reference){
	angle = (M_PI * angle) / 180;
	double cos_angle = cos(angle); double sin_angle = sin(angle);

	Matrix<double> *matrix = new Matrix<double>(4,4);
	switch(axis){
	case MatrixFactory::X:
		(*matrix)(0,0) = 1; 	(*matrix)(0,1) = 0; 			(*matrix)(0,2) = 0; 			(*matrix)(0,3) = 0;
		(*matrix)(1,0) = 0; 	(*matrix)(1,1) = cos_angle; 	(*matrix)(1,2) = -sin_angle; 	(*matrix)(1,3) = reference.y*(1 - cos_angle) +(reference.z*sin_angle);
		(*matrix)(2,0) = 0; 	(*matrix)(2,1) = sin_angle; 	(*matrix)(2,2) = cos_angle; 	(*matrix)(2,3) = reference.z*(1 - cos_angle) -(reference.y*sin_angle);
		(*matrix)(3,0) = 0; 	(*matrix)(3,1) = 0; 			(*matrix)(3,2) = 0; 			(*matrix)(3,3) = 1;
		break;
	case MatrixFactory::Y:
		(*matrix)(0,0) = cos_angle; 	(*matrix)(0,1) = 0; 	(*matrix)(0,2) = sin_angle; 	(*matrix)(0,3) = reference.x*(1 - cos_angle) -(reference.z*sin_angle);
		(*matrix)(1,0) = 0; 			(*matrix)(1,1) = 1; 	(*matrix)(1,2) = 0; 			(*matrix)(1,3) = 0;
		(*matrix)(2,0) = -sin_angle; 	(*matrix)(2,1) = 0; 	(*matrix)(2,2) = cos_angle; 	(*matrix)(2,3) = reference.z*(1 - cos_angle) +(reference.x*sin_angle);
		(*matrix)(3,0) = 0; 			(*matrix)(3,1) = 0; 	(*matrix)(3,2) = 0; 			(*matrix)(3,3) = 1;
		break;
	case MatrixFactory::Z:
		(*matrix)(0,0) = cos_angle; 	(*matrix)(0,1) = -sin_angle; 	(*matrix)(0,2) = 0; 	(*matrix)(0,3) = reference.x*(1 - cos_angle) +(reference.y*sin_angle);
		(*matrix)(1,0) = sin_angle; 	(*matrix)(1,1) = cos_angle; 	(*matrix)(1,2) = 0; 	(*matrix)(1,3) = reference.y*(1 - cos_angle) -(reference.x*sin_angle);
		(*matrix)(2,0) = 0; 			(*matrix)(2,1) = 0; 			(*matrix)(2,2) = 1; 	(*matrix)(2,3) = 0;
		(*matrix)(3,0) = 0; 			(*matrix)(3,1) = 0; 			(*matrix)(3,2) = 0; 	(*matrix)(3,3) = 1;
		break;
	}

	return matrix;
}

Matrix<double>* MatrixFactory::bSplineMatrix(){
	Matrix<double> * matrix = new Matrix<double>(4,4);

	(*matrix)(0,0) = -1.0/6; 	(*matrix)(0,1) = 1.0/2; 	(*matrix)(0,2) = -1.0/2 ; 	(*matrix)(0,3) = 1.0/6;
	(*matrix)(1,0) = 1.0/2;  	(*matrix)(1,1) = -1;    	(*matrix)(1,2) = 1.0/2 ;  	(*matrix)(1,3) = 0;
	(*matrix)(2,0) = -1.0/2;  	(*matrix)(2,1) = 0;     	(*matrix)(2,2) = 1.0/2 ;  	(*matrix)(2,3) = 0;
	(*matrix)(3,0) = 1.0/6;  	(*matrix)(3,1) = 4.0/6; 	(*matrix)(3,2) = 1.0/6;   	(*matrix)(3,3) = 0;

	return matrix;
}

Matrix<double>* MatrixFactory::bezierMatrix(){
	Matrix<double> * matrix = new Matrix<double>(4,4);

	(*matrix)(0,0) = -1; 		(*matrix)(0,1) = 3; 		(*matrix)(0,2) = -3 ; 		(*matrix)(0,3) = 1;
	(*matrix)(1,0) = 3;  		(*matrix)(1,1) = -6;    	(*matrix)(1,2) = 3 ;  		(*matrix)(1,3) = 0;
	(*matrix)(2,0) = -3;  		(*matrix)(2,1) = 3;     	(*matrix)(2,2) = 0 ;  		(*matrix)(2,3) = 0;
	(*matrix)(3,0) = 1;  		(*matrix)(3,1) = 0; 		(*matrix)(3,2) = 0;   		(*matrix)(3,3) = 0;

	return matrix;
}

Matrix<Coordinate>* MatrixFactory::surfaceGeoMatrix(const Matrix<double> *points, unsigned geoNumber){
	Matrix<Coordinate>* matrix = new Matrix<Coordinate>(4,4);
	int coordinateNumber = 16*geoNumber;

	for(unsigned i = 0; i < (*matrix).getRows(); i++){
		for(unsigned j = 0; j < (*matrix).getColumns(); j++){
			(*matrix)(i,j) = Coordinate((*points)(0,coordinateNumber),
					                    (*points)(1,coordinateNumber),
					                    (*points)(2,coordinateNumber)
					                    );
			coordinateNumber++;
		}
	}

	return matrix;
}
