#include "Polygon.h"

Polygon::Polygon(CoordinateList& points, const std::string& name, bool filled, const DrawCfg& drawCfg)
	: Polyline(points, name, drawCfg) {
	if(filled)
		type = FILLED_POLYGON;
	else
		type = POLYGON;
}

Polygon::Polygon(const Polygon& other) : Polyline(other){
	type = other.type;
}

GeometricElement* Polygon::clone() const{
	return new Polygon(*this);
}

Coordinate Polygon::first() const{
	return Coordinate((*matrix)(0,0), (*matrix)(1,0), (*matrix)(2,0));
}

Coordinate Polygon::last() const{
	unsigned last = nPoints() - 1;
	return Coordinate((*matrix)(0,last), (*matrix)(1,last), (*matrix)(2,last));
}

//reference: Determining whether or not a polygon (2D) has its
//vertices ordered clockwise or counterclockwise.
//acessible in: http://debian.fmi.uni-sofia.bg/~sergei/cgsr/docs/clockwise.htm
bool Polygon::isClockwise() const{
	int counter = 0;

	unsigned nrPoints = nPoints();
	for(unsigned i = 0; i < nrPoints; i++){
		//feel free to implement the special cases outside the for to avoid the calculations below
		unsigned j = (i + 1) % nrPoints;
		unsigned k = (i + 2) % nrPoints;

		double crossMagnitude = ( (*matrix)(0,j) - (*matrix)(0,i))
				* ( (*matrix)(1,k) - (*matrix)(1,j));
		crossMagnitude -= ( ( (*matrix)(1,j) - (*matrix)(1,i) )
				* ( (*matrix)(0,k) - (*matrix)(0,j)) );

		if(crossMagnitude < 0)
			counter--;
		if(crossMagnitude > 0)
			counter++;
	}

	return (counter < 0);
}

CoordinateList Polygon::getCoordinatesCW() const{
	CoordinateList coordinates(nPoints());

	if(isClockwise()){
		for(unsigned i = 0; i < nPoints(); i++)
			coordinates[i] = Coordinate( (*matrix)(0,i), (*matrix)(1,i), (*matrix)(2,i));
	}else{
		for(int i = nPoints() - 1; i >= 0; i--)
			coordinates[i] = Coordinate( (*matrix)(0,i), (*matrix)(1,i), (*matrix)(2,i));
	}

	return coordinates;
}


LineList Polygon::getLines() const{
	LineList lines = Polyline::getLines();
	lines.push_back(Line(last(), first(), std::string(name).append("###last"), drawingConfig));
	return lines;
}


ElementsList Polygon::clip(const Rectangle& clippingArea) const{
	if(type == GeometricElement::POLYGON)
		return Polyline::clip(clippingArea);
	else
		return clipFilled(clippingArea);
}

void Polygon::waGetIntersectionPair(const Rectangle& clippingArea, IntersectionList& subj, IntersectionList& clip) const{
	//retrieves coordinates
    CoordinateList subjCw = getCoordinatesCW();
    CoordinateList clipCw = clippingArea.getCoordinatesCW();

    subjCw.push_back(subjCw[0]);
    clipCw.push_back(clipCw[0]);

    CoordinateList::iterator i = subjCw.begin();
    CoordinateList::iterator j = clipCw.begin();

    for(unsigned lineCount = 0; lineCount < subjCw.size() -1; lineCount++){
    	Line lineSubj(subjCw[lineCount], subjCw[lineCount + 1], "###line", drawingConfig);
		j = clipCw.begin();
		subj.push_back(std::pair<bool, Coordinate>(false, (*i)));

		for(unsigned clipCount = 0; clipCount < clipCw.size() - 1; clipCount++){
			Line lineClip(clipCw[clipCount], clipCw[clipCount + 1], "###line", drawingConfig);
			Coordinate *intersection = lineSubj.intersects(lineClip);
			if(intersection){
				subj.push_back(std::pair<bool, Coordinate>(true, *intersection));
				delete intersection;
			}

			j++;
		}

		i++;
    }


    j = clipCw.begin();
    for(unsigned clipCount = 0; clipCount < clipCw.size() -1; clipCount++){
        Line lineClip(clipCw[clipCount], clipCw[clipCount + 1], "###line", drawingConfig);
        i = subjCw.begin();
        clip.push_back(std::pair<bool, Coordinate>(false, (*j)));

        for(unsigned lineCount = 0; lineCount < subjCw.size() -1; lineCount++){
        	Line lineSubj(subjCw[lineCount], subjCw[lineCount + 1], "###line", drawingConfig);
			Coordinate *intersection = lineClip.intersects(lineSubj);
			if(intersection){
				clip.push_back(std::pair<bool, Coordinate>(true, *intersection));
				delete intersection;
			}

			i++;
        }

        j++;
    }
}

//returns false if there are no points outside
void Polygon::waFindEnteringIntersections(IntersectionList& subj, const Rectangle& clippingArea, CoordinateList& enteringIntersections) const{
	for(IntersectionList::iterator i = subj.begin() + 1; i != subj.end(); i++){
		if((*i).first){
			IntersectionList::iterator next = i + 1;
			if(next == subj.end())
				next = subj.begin();

			Point nextPoint((*next).second,"nextPoint");

			if((nextPoint.clip(clippingArea).size() > 0))
				enteringIntersections.push_back((*i).second);
		}
	}
}

Polygon::IntersectionList::iterator Polygon::waFindNearest(IntersectionList& list, Coordinate coord) const{
	double distance = (*list.begin()).second.distance(coord);
	IntersectionList::iterator nearest = list.begin();
	for(IntersectionList::iterator i = ++(list.begin()); i != list.end(); i++){
		double thisDistance = (*i).second.distance(coord);
		if(thisDistance < distance){
			distance = thisDistance;
			nearest = i;
		}
	}

	return nearest;
}

ElementsList Polygon::waCutPolygons(IntersectionList& subj, IntersectionList& clip, CoordinateList& enteringIntersections) const{
	ElementsList polygons;

	for(unsigned i = 0; i < enteringIntersections.size(); i++){
		CoordinateList polygonCoordinates;

		IntersectionList *currentList = &subj;
		IntersectionList *otherList = &clip;
		IntersectionList::iterator subjIntersection = waFindNearest(*currentList, enteringIntersections[i]);
		IntersectionList::iterator clipIntersection = waFindNearest(*otherList, enteringIntersections[i]);

		polygonCoordinates.push_back((*subjIntersection).second);
		for(IntersectionList::iterator j = subjIntersection +1; j != subjIntersection && j != clipIntersection; j++){
			if(j == currentList->end())
				j = currentList->begin();

			polygonCoordinates.push_back((*j).second);

			if((*j).first){
				j = waFindNearest(*otherList, (*j).second);

				IntersectionList *swapList = currentList;
				currentList = otherList;
				otherList = swapList;
			}
		}
	}

	return polygons;
}

//fixme: doesn't work
ElementsList Polygon::clipFilled(const Rectangle& clippingArea) const{
	ElementsList polygons;
	polygons.push_back(this->clone());
	return polygons;
//	IntersectionList subj; IntersectionList clip;
//	waGetIntersectionPair(clippingArea, subj, clip);
//
//	CoordinateList enteringIntersections;
//	waFindEnteringIntersections(subj, clippingArea, enteringIntersections);
//
//	if(enteringIntersections.size() == 0){
//		Point point(first(),"point");
//
//		ElementsList clipped = point.clip(clippingArea);
//		if(clipped.size() > 0){
//			delete (*clipped.begin());
//
//			ElementsList polygons;
//			polygons.push_back(this->clone());
//			return polygons;
//		}
//	}
//
//	return waCutPolygons(subj, clip, enteringIntersections);
}

Polygon::~Polygon() {
}

// ---- Rectangle ----

Rectangle::Rectangle(const Coordinate& min, const Coordinate& max, const std::string& name,
		bool filled, const DrawCfg& drawCfg) :
		Polygon(*createCoordinates(min,max),name,filled,drawCfg)
	{

}

Coordinate Rectangle::min() const{
	return Coordinate((*matrix)(0,0),(*matrix)(1,0),(*matrix)(2,0));
}

Coordinate Rectangle::max() const{
	return Coordinate((*matrix)(0,2),(*matrix)(1,2),(*matrix)(2,2));
}

double Rectangle::height() const{
	return (*matrix)(1,2) - (*matrix)(1,0);
}

double Rectangle::width() const{
	return (*matrix)(0,2) - (*matrix)(0,0);
}

std::auto_ptr<CoordinateList> Rectangle::createCoordinates(const Coordinate& min, const Coordinate& max){
	CoordinateList *coordinates = new CoordinateList;
	coordinates->push_back(min);
	coordinates->push_back(Coordinate(min.x, max.y,max.z));
	coordinates->push_back(max);
	coordinates->push_back(Coordinate(max.x,min.y,max.z));

	return std::auto_ptr<CoordinateList>(coordinates);
}

TiXmlElement* Polygon::toXmlElement() const{
	TiXmlElement *polygonElement = new TiXmlElement("poligono");

	polygonElement->SetAttribute("nome",name.c_str());
	polygonElement->SetAttribute("cor", drawingConfig.getColor().colorCode().c_str());
	if(is(GeometricElement::FILLED_POLYGON))
		polygonElement->SetAttribute("cheio", "verdadeiro");
	else
		polygonElement->SetAttribute("cheio", "falso");

	for(unsigned i = 0; i < nPoints(); i++){
		TiXmlElement *point = new TiXmlElement("ponto");
		point->SetDoubleAttribute("x",(*matrix)(0,i));
		point->SetDoubleAttribute("y",(*matrix)(1,i));
		point->SetDoubleAttribute("z",(*matrix)(2,i));

		polygonElement->LinkEndChild(point);
	}

	return polygonElement;
}

Polygon* Polygon::createFromXml(TiXmlElement* element){
	if(!element)
		throw InvalidParameterException("invalid XML.");

	CoordinateList coordinates;
	for(TiXmlElement *i = element->FirstChildElement("ponto"); i != 0; i = i->NextSiblingElement("ponto")){
		Coordinate coordinate;
		i->QueryDoubleAttribute("x",&coordinate.x);
		i->QueryDoubleAttribute("y",&coordinate.y);
		i->QueryDoubleAttribute("z",&coordinate.z);

		coordinates.push_back(coordinate);
	}

	std::string name(element->Attribute("nome"));
	DrawCfg drawingCfg(Color(std::string("#").append(element->Attribute("cor"))));

	std::string cheioStr(element->Attribute("cheio"));
	bool cheio = true;
	if(cheioStr == "falso")
		cheio = false;

	return new Polygon(coordinates,name,cheio,drawingCfg);
}
