#include "Polyline.h"

Polyline::Polyline(const CoordinateList& points, const std::string& name, const DrawCfg& drawCfg)
	: GeometricElement(points.size(), GeometricElement::POLYLINE, name, drawCfg)
{
	if(nPoints() < 2)
		throw InvalidParameterException("polyline with less than 2 points");

	for(unsigned j = 0; j < nPoints(); j++){
		(*matrix)(0,j) = points[j].x;
		(*matrix)(1,j) = points[j].y;
		(*matrix)(2,j) = points[j].z;
		(*matrix)(3,j) = 1;
	}

	recalculateCenter();
}

Polyline::Polyline(const Polyline& other) : GeometricElement(other){
}

GeometricElement* Polyline::clone() const{
	return new Polyline(*this);
}

LineList Polyline::getLines() const{
	LineList lines;

	for(unsigned i = 0; i < nPoints() - 1; i++){
		Coordinate first(Coordinate((*matrix)(0,i),(*matrix)(1,i),(*matrix)(2,i)));
		Coordinate second(Coordinate((*matrix)(0,i+1),(*matrix)(1,i+1),(*matrix)(2,i+1)));

		std::stringstream buffer;
		buffer << "#" << i;
		lines.push_back(Line(first, second, getName().append(buffer.str()), getDrawingConfig()));
	}

	return lines;
}

ElementsList Polyline::clip(const Rectangle& clippingArea) const{
	LineList lines = getLines();
	ElementsList list;

	for(LineList::iterator i = lines.begin(); i != lines.end(); i++){
		ElementsList elements = (*i).clip(clippingArea);
		if(elements.size() > 0)
			list.push_back(*(elements.begin()));
	}

	return list;
}

ElementsList Polyline::clip3d(const Rectangle& clippingArea, const Coordinate& projectionCenter) const{
	LineList lines = getLines();
	ElementsList list;

	for(LineList::iterator i = lines.begin(); i != lines.end(); i++){
		ElementsList elements = (*i).clip3d(clippingArea,projectionCenter);
		if(elements.size() > 0)
			list.push_back(*(elements.begin()));
	}

	return list;
}

TiXmlElement* Polyline::toXmlElement() const{
	TiXmlElement *polylineElement = new TiXmlElement("polilinha");

	polylineElement->SetAttribute("nome",name.c_str());
	polylineElement->SetAttribute("cor", drawingConfig.getColor().colorCode().c_str());

	for(unsigned i = 0; i < nPoints(); i++){
		TiXmlElement *point = new TiXmlElement("ponto");
		point->SetDoubleAttribute("x",(*matrix)(0,i));
		point->SetDoubleAttribute("y",(*matrix)(1,i));
		point->SetDoubleAttribute("z",(*matrix)(2,i));

		polylineElement->LinkEndChild(point);
	}

	return polylineElement;
}

Polyline* Polyline::createFromXml(TiXmlElement* element){
	if(!element)
		throw InvalidParameterException("invalid XML.");

	CoordinateList coordinates;
	for(TiXmlElement *i = element->FirstChildElement("ponto"); i != 0; i = i->NextSiblingElement("ponto")){
		Coordinate coordinate;
		i->QueryDoubleAttribute("x",&coordinate.x);
		i->QueryDoubleAttribute("y",&coordinate.y);
		i->QueryDoubleAttribute("z",&coordinate.z);

		coordinates.push_back(coordinate);
	}

	std::string name(element->Attribute("nome"));
	DrawCfg drawingCfg(Color(std::string("#").append(element->Attribute("cor"))));

	return new Polyline(coordinates,name,drawingCfg);
}

Polyline::~Polyline() {}
