#include "Window.h"
#include "Viewport.h"
#include "Point.h"

#define GS_PERSPECTIVE_PROJECTION
//#define GS_CLIPPING_3D

Window::Window(const Coordinate& minCoord, const Coordinate& maxCoord, const Coordinate& projectionCenter)
	: wRectangle(minCoord, maxCoord,"windowPolygon", false, DrawCfg(Color("#FF0000"), 2)),
	  projectionCenter(projectionCenter),
	  vpTMatrix(4,4)  {
}

Window::Window(const Window& other) :
		wRectangle(other.wRectangle), projectionCenter(other.projectionCenter),
		vpTMatrix(4,4){
}

//Reference:  Foley, James D., Computer Graphics: Principles and pratices, 2nd Ed, page 212.
void Window::setViewportMatrix(const Viewport& viewport){
	double xRatio = viewport.getWidth() / wRectangle.width();
	double yRatio = viewport.getHeight() / wRectangle.height();

	Coordinate minimum = wRectangle.min();

	vpTMatrix(0,0) = xRatio; 	vpTMatrix(0,1) = 0; 		vpTMatrix(0,2) = 0;		vpTMatrix(0,3) = -1 * minimum.x * xRatio;
	vpTMatrix(1,0) = 0; 		vpTMatrix(1,1) = yRatio; 	vpTMatrix(1,2) = 0;		vpTMatrix(1,3) = -1 * minimum.y * yRatio;
	vpTMatrix(2,0) = 0; 		vpTMatrix(2,1) = 0; 		vpTMatrix(2,2) = 0;		vpTMatrix(2,3) = 0;
	vpTMatrix(3,0) = 0; 		vpTMatrix(3,1) = 0; 		vpTMatrix(3,2) = 0;		vpTMatrix(3,3) = 0;
}

void Window::draw(DisplayFile& displayFile, Viewport& viewport){
	setViewportMatrix(viewport);

	#ifndef GS_PERSPECTIVE_PROJECTION
		DisplayFile *clippedDisplay = clip(displayFile);
	#else
		DisplayFile *projectedDisplay = project(displayFile);
		DisplayFile *clippedDisplay = clip(*projectedDisplay);
	#endif

	for(DisplayFile::Iterator i = clippedDisplay->begin(); i != clippedDisplay->end(); i++){
		GeometricElement *transformedElement = (*i)->clone();
		transformedElement->transform(vpTMatrix);
		viewport.draw(*transformedElement);
		delete transformedElement;
	}

	delete clippedDisplay;

	#ifdef GS_PERSPECTIVE_PROJECTION
		delete projectedDisplay;
	#endif
}

DisplayFile* Window::project(DisplayFile& displayFile){
	DisplayFile* projectedDisplay = new DisplayFile();

	for(DisplayFile::Iterator i = displayFile.begin(); i != displayFile.end(); i++){
		GeometricElement* clone = (*i)->clone();
		clone->project(wRectangle,projectionCenter);

		#ifdef GS_CLIPPING_3D
			ElementsList list = clone->clip3d(wRectangle,projectionCenter);
			projectedDisplay->add(list);
		#else
			projectedDisplay->add(clone);
		#endif
	}

	return projectedDisplay;
}

DisplayFile* Window::clip(DisplayFile& displayFile){
	DisplayFile *clippedDisplay = new DisplayFile();

	Rectangle clippingArea(wRectangle);
	clippingArea.scale(SUBCANVAS_RATIO, SUBCANVAS_RATIO, 1);

	for(DisplayFile::Iterator i = displayFile.begin(); i != displayFile.end(); i++ ){
		ElementsList clippedElement = (*i)->clip(clippingArea);
		clippedDisplay->add(clippedElement);
	}

	ElementsList subcanvas = clippingArea.clip(wRectangle);
	clippedDisplay->add(subcanvas);

	return clippedDisplay;
}

void Window::scale(double xScale, double yScale, double zScale){
	wRectangle.scale(xScale,yScale, zScale);
}

void Window::translate(double xOffset, double yOffset, double zOffset){
	wRectangle.translate(xOffset, yOffset, zOffset);
}

Coordinate Window::getCenter() const{
	return wRectangle.getCenter();
}

Coordinate Window::getProjectionCenter() const{
	return projectionCenter;
}

Coordinate Window::min() const {
	return wRectangle.min();
}

Coordinate Window::max() const{
	return wRectangle.max();
}


Window::~Window() {
}
