#ifndef GEOMETRICELEMENTTESTS_H_
#define GEOMETRICELEMENTTESTS_H_

#include <cpptest.h>
#include "Point.h"
#include "Line.h"
#include "Polyline.h"
#include "Polygon.h"
#include "Vector.h"

class GeometricElementTests : public Test::Suite {
public:
	GeometricElementTests(){
		TEST_ADD(GeometricElementTests::create_points);
		TEST_ADD(GeometricElementTests::create_lines);
		TEST_ADD(GeometricElementTests::create_polylines);
		TEST_ADD(GeometricElementTests::create_polygons);
		TEST_ADD(GeometricElementTests::create_coordinates);
		TEST_ADD(GeometricElementTests::create_elements_with_name);
		TEST_ADD(GeometricElementTests::polygon_orientation);
		TEST_ADD(GeometricElementTests::line_intersection);
	}

	void create_points(){
		Point point(Vector(130,10,1),"q");

		TEST_ASSERT(point.getCoordinate() == Coordinate(130,10,1));

		Point point2(point);
		TEST_ASSERT(point == point2);

		Point point3(Vector(0,0,0),"a");
		TEST_ASSERT(point != point3);
	}

	void create_lines(){
		Point point1(Vector(10,10,0),"a");
		Point point2(Vector(0,0,0),"a");
		Point point3(Vector(5,5,0),"a");
		Point point4(Vector(20,3,0),"a");
		Line line(Vector(10,10,0),Vector(0,0,0),"a");
		Line line2(Vector(10,10,0),Vector(5,5,0),"a");

		Matrix<double> lineMatrix(3,2);
		lineMatrix(0,0) = 10;  lineMatrix(0,1) = 0;
		lineMatrix(1,0) = 10;  lineMatrix(1,1) = 0;
		lineMatrix(2,0) = 0;	lineMatrix(2,1) = 0;


		TEST_ASSERT(line != line2);
		TEST_ASSERT(line2 != point1);
	}

	void create_polylines(){
		CoordinateList pointsList;
		CoordinateList pointsList2;
		for(unsigned i = 0; i < 6; i++){
			pointsList.push_back(Vector(i*10,i*10,i*10));
			pointsList2.push_back(Vector(i*20,i*20,i*20));
		}

		Matrix<double> polylineMatrix(3,6);
		Matrix<double> polylineMatrix2(3,6);
		for(unsigned i = 0; i < 6; i++){
			for(unsigned j = 0; j < 3; j++){
				polylineMatrix(j,i) = i*10;
				polylineMatrix2(j,i) = i*20;
			}
		}

		Polyline polyline1(pointsList,"a");
		Polyline polyline2(pointsList2,"a");
		Polyline polyline3(polyline1);

		TEST_ASSERT(polyline1 != polyline2);
		TEST_ASSERT(polyline3 == polyline1);
		TEST_ASSERT(polyline3 != polyline2);
	}

	void create_polygons(){
		CoordinateList pointsList;
		CoordinateList pointsList2;
		CoordinateList pointsList3;
		for(unsigned i = 0; i < 6; i++){
			pointsList.push_back(Vector(i*10,i*10,i*10));
			pointsList2.push_back(Vector(i*20,i*20,i*10));
			pointsList3.push_back(Vector(i*20,i*20,i*10));
		}
		pointsList3.push_back(Vector(0,0,0));

		Polyline polyline1(pointsList,"a");
		Polyline polyline2(pointsList2,"a");
		Polyline polyline3(pointsList3,"a");

		Polygon polygon(pointsList2,"a",false);

		TEST_ASSERT(polygon != polyline1);
	}

	void create_coordinates(){
		Vector coordinate1(0,0,1);
		TEST_ASSERT(coordinate1.x == 0 && coordinate1.y == 0 && coordinate1.z == 1);

		Vector coordinate2(10,100,1);
		TEST_ASSERT(coordinate2.x == 10 && coordinate2.y == 100 && coordinate2.z == 1);

		Vector coordinate3(-43.43, 66.23, 1);
		TEST_ASSERT(coordinate3.x == -43.43 && coordinate3.y == 66.23 && coordinate3.z == 1);
	}

	void create_elements_with_name(){
		Vector coord1(10,10,10);
		Vector coord2(11,10,10);
		Vector coord3(10,23,10);
		Vector coord4(10,12,10);
		Vector coord5(44,10,10);

		Point p1(coord1,"p1");
		Point p2(coord2,"p2");
		Point p3(p1);
		Line l1(coord1,coord2,"l1");
		Line l2(coord4,coord2,"l2");

		TEST_ASSERT(p1.getName() == "p1");
		TEST_ASSERT(p2.getName() == "p2");
		TEST_ASSERT(p3.getName() == "p1");

		CoordinateList coordlist;
		coordlist.push_back(coord1);
		coordlist.push_back(coord2);
		coordlist.push_back(coord3);
		coordlist.push_back(coord4);
		coordlist.push_back(coord5);

		Polyline poly1(coordlist,"p1");

		Polygon plg1(coordlist,"plg1",false);
		Polygon plg2(plg1);

		TEST_ASSERT(plg1.getName() == "plg1");
		TEST_ASSERT(plg2.getName() == "plg1");

		Line linec(Coordinate(0,0,1),Coordinate(500,500,1),"linec");
		TEST_ASSERT(linec.getCenter() == Coordinate(250,250,1));

	}

	void polygon_orientation(){
		CoordinateList coordinates1;

		coordinates1.push_back(Coordinate(0,0,1));
		coordinates1.push_back(Coordinate(0,100,1));
		coordinates1.push_back(Coordinate(100,100,1));
		coordinates1.push_back(Coordinate(100,0,1));

		Polygon polygon1(coordinates1,"p1",false);

		TEST_ASSERT(polygon1.isClockwise());


		CoordinateList coordinates2;
		coordinates2.push_back(Coordinate(0,0,1));
		coordinates2.push_back(Coordinate(100,0,1));
		coordinates2.push_back(Coordinate(100,100,1));
		coordinates2.push_back(Coordinate(0,100,1));
		Polygon polygon2(coordinates2,"p2",false);

		TEST_ASSERT(!polygon2.isClockwise());
	}

#define notInt(x,y) TEST_ASSERT(!x.intersects(y));

#define yesInt(z,k) TEST_ASSERT(z.intersects(k)); \
	if(z.intersects(k)) std::cout << z.intersects(k)->x << ", " << z.intersects(k)->y << std::endl;

	void line_intersection(){
		Line l1(Coordinate(-500,500,1),Coordinate(-500,-500,1),"l1");
		Line l2(Coordinate(-480,480,1),Coordinate(450,-480,1),"l2");
		Line l3(Coordinate(-450,-300,1),Coordinate(550,-300,1),"l3");
		Line l4(Coordinate(-400,-400,1),Coordinate(400,500,1),"l4");
		Line l5(Coordinate(-400,500,1),Coordinate(550,-350,1),"l5");
		Line l6(Coordinate(-300,400,1),Coordinate(550,400,1),"l6");
		Line l7(Coordinate(500,490,1),Coordinate(500,-500,1),"l7");

		std::cout << std::endl << std::endl << "l1" << std::endl;
		notInt(l1,l2);
		notInt(l1,l3);
		notInt(l1,l4);
		notInt(l1,l5);
		notInt(l1,l6);
		notInt(l1,l7);

		std::cout << std::endl << std::endl << "l2" << std::endl;

		notInt(l2,l1);
		yesInt(l2,l3);
		yesInt(l2,l4);
		notInt(l2,l5);
		notInt(l2,l6);
		notInt(l2,l7);

		std::cout << std::endl << std::endl << "l3" << std::endl;

		notInt(l3,l1);
		yesInt(l3,l2);
		yesInt(l3,l4);
		yesInt(l3,l5);
		notInt(l3,l6);
		yesInt(l3,l7);

		std::cout << std::endl << std::endl << "l4" << std::endl;

		notInt(l4,l1);
		yesInt(l4,l2);
		yesInt(l4,l3);
		yesInt(l4,l5);
		yesInt(l4,l6);
		notInt(l4,l7);

		std::cout << std::endl << std::endl << "l5" << std::endl;

		notInt(l5,l1);
		notInt(l5,l2);
		yesInt(l5,l3);
		yesInt(l5,l4);
		yesInt(l5,l6);
		yesInt(l5,l7);

		std::cout << std::endl << std::endl << "l6" << std::endl;

		notInt(l6,l1);
		notInt(l6,l2);
		notInt(l6,l3);
		yesInt(l6,l4);
		yesInt(l6,l5);
		yesInt(l6,l7);

		std::cout << std::endl << std::endl << "l7" << std::endl;

		notInt(l7,l1);
		notInt(l7,l2);
		yesInt(l7,l3);
		notInt(l7,l4);
		yesInt(l7,l5);
		yesInt(l7,l6);
	}

	virtual ~GeometricElementTests() {}
};

#endif /* GEOMETRICELEMENTTESTS_H_ */
